/* $Id: layout.c,v 1.4 2002/02/09 23:07:05 mquinson Exp $ */

/* layout - compute where to put each machines in the visualization space    */
/*          see header file for the algorithm                                */
/* Author: Martin Quinson                                                    */
/* License: LGPL                                                             */


#include <stdio.h>       /* printf() */
#include <stdlib.h>      /* atoi() */
#include <string.h>      /* string functions */
#include <strings.h>     /* strncasecmp() on aix */
#include <unistd.h>      /* getopt() */
#include <math.h>        /* sin, cos */

#include "nws_api.h"     /* NWS programming interface */
#include "diagnostic.h"  /* ABORT() */

#include "jval.h"
#include "jrb.h"

#include "machines_cliques.h" /* machines and cliques sets */
#include "settings.h"         /* constants influing the layout computation */
#include "layout.h"


/* Level type */
typedef struct {
  double level; /* number of this level (0 is bottom) */
  JRB clique;   /* all cliques present on this level */
  JRB machine;  /* all machines of this level (not their name, the machines) */
} layout_level_t;

static layout_level_t *lvl=NULL;
static int lvl_count=0;

/*----[ level_dump ]---------------------------------------------------------*/
/* debugging is a Good Thing (TM)                                            */
/*---------------------------------------------------------------------------*/
static void level_dump() {
  int lvl_cur,i;
  JRB sc,sm;
  clique_t *c;
  machine_t *m;
  char *name;
  int first; /* to put "," between elements */
  
  for (lvl_cur=lvl_count-1; lvl_cur>=0; lvl_cur--) {

    printf ("----[ LEVEL %d ]-------------------\n",lvl_cur);
    printf (" Cliques: ");
    jrb_traverse(sc,lvl[lvl_cur].clique) {
      c = (clique_t *)jval_v(sc->val);
      if (!c) {
	ERROR("Damnit! Clique not found during traversal\n");
	return;
      }
      printf("%s=(",c->name);
      first=1;
      for (i=0 ; i<c->count ; i++) {
	if (first) {
	  first=0;
	} else {
	  printf(",");
	}
	printf("%s",c->machine[i]->fqdn);
      }
      printf (")\n          ");
    }
    printf ("\n Machines: ");
    jrb_traverse(sm,lvl[lvl_cur].machine) {
      m=(machine_t*)jval_v(sm->val);
      if (!sm) {
	ERROR("Damnit ! unknown machine during traverse...\n");
	return;
      }

      if (m->z==lvl_cur)
	printf("%s pos=(%.0f,%.0f,%.0f);weight=%d;memberof=%d\n           ",
	       name,m->x,m->y,m->z,
	       m->weight,m->memberof);
    }    
    printf("\n");
  }
  printf ("----[ GROUND ]-------------------\n\n\n");
}

/*----[ level_free ]---------------------------------------------------------*/
/* Memory leak is a Bad Thing (TM)                                           */
/*---------------------------------------------------------------------------*/
static void level_free() {
  int lvl_cur;
  /*  JRB sc; FIXME:
      char *name;*/

  if (!lvl) return;
  for (lvl_cur=0; lvl_cur<lvl_count; lvl_cur++) {  
    //    jrb_traverse(sc,lvl[lvl_cur].clique_names) {
    //      name=(char*)jval_v(sc->val);
    //      free(name);
    //    }

    jrb_free_tree(lvl[lvl_cur].clique);
    jrb_free_tree(lvl[lvl_cur].machine);
  }
  free(lvl);
  lvl=NULL;
  lvl_count=0;
}

/*----[ layout_phase1 ]------------------------------------------------------*/
/* phase 1: count in how many clique are each nodes                          */
/*---------------------------------------------------------------------------*/
static int layout_phase1() {
  JRB sm,sc; /* search {machine,clique} */
  machine_t *m;
  clique_t *c;
  int i;

  DDEBUG("Layout phase 1\n");
  /* first cleanup stuff from previous runs. */
  jrb_traverse(sm,machines) {
    m=(machine_t*)jval_v(sm->val);
    m->memberof = 0;
    m->father = NULL;
  }

  /* do the job */
  jrb_traverse(sc,cliques) {
    c=(clique_t*)jval_v(sc->val);
    for (i=0; i<c->count; i++)
      c->machine[i]->memberof++;
  }

  /* FIXME: only one tree allowed => standalone machines are removed */
  /* causes segfault, so removed :)
  jrb_traverse(sm,machines) {
    m=(machine_t*)jval_v(sm->val);
    if (m->fqdn[0] != '\0' && m->memberof == 0) {
      DDEBUG1("%s is in no clique. Remove it.\n",m->fqdn);
      jrb_delete_node(sm); 
      free(m);
      DDEBUG("Done\n");
    }
  }
  */

  return 1;
}

/*---------------------------------------------------------------------------*/
/* phase 2: figure out clique level                                          */
/*   Level of each clique = average of the memberof of its members           */
/*   Renumerate the layer to be a suite of int (1.3 ; 3 ; 3.4 -> 1 ; 2 ; 3)  */
/*   Merge adjacent layers containing cliques with no shared hosts           */
/*---------------------------------------------------------------------------*/

/* Temporary type to sort the cliques depending on their level using qsort */
typedef struct {
  double level;
  clique_t *clique;
} layout_phase2_cell_t;
/* Corresponding cmp function */
static int layout_phase2_cell_cmp(const void *a, const void *b){
  layout_phase2_cell_t *x=(layout_phase2_cell_t*)a;
  layout_phase2_cell_t *y=(layout_phase2_cell_t*)b;
  if (x->level == y->level)
    return 0;
  if (x->level > y->level)
    return 1;
  return -1;
}

/*----[ layout_phase2 ]------------------------------------------------------*/
/* compute along the Z axis                                                  */
/*---------------------------------------------------------------------------*/
static int layout_phase2() {
  JRB sm,sc; /* search machines & clique */
  machine_t *m;
  clique_t *c;
  int count,lvl_cur, i,j;

  layout_phase2_cell_t *clique_set=NULL;
  int clique_count=0;
  char *name;

  DDEBUG("Layout phase 2.\n");
  WARN("FIXME: redo that stuff without layout_phase2_cell_t\n");
  DDEBUG("  Step 1 & 2: initialization stuff\n");
  /* Step 1: level of each clique = average memberof of its members     */
  jrb_traverse(sc,cliques) {
    if (clique_count++) {
      clique_set=(layout_phase2_cell_t*)realloc(clique_set,
					     sizeof(layout_phase2_cell_t)*
					     clique_count);
    } else {
      clique_set=(layout_phase2_cell_t*)malloc(sizeof(layout_phase2_cell_t));
    }
    if (!clique_set)
      FAIL("Out of memory\n");

    clique_set[ clique_count-1 ].level=0;
    clique_set[ clique_count-1 ].clique=(clique_t*)jval_v(sc->val);

    count=0;

    /* count the machines in this clique */
    for (i=0; i<clique_set[ clique_count-1 ].clique->count; i++) {
      clique_set[ clique_count-1 ].level +=
	clique_set[ clique_count-1 ].clique->machine[i]->memberof;
      count++;
    }
    clique_set[ clique_count-1 ].level /= count;
    DDEBUG2("Clique %s is at level %.2f\n",
	    clique_set[ clique_count-1].clique->name,
	    clique_set[ clique_count-1 ].level);
  }

  if (!clique_set)
    FAIL("No clique found. This case is not (yet) handled by the layout stuff.");

  /* Step 2: Renumerate the level to be a suite of int */
  /*         (1.3 ; 3 ; 3.5-> 1 ; 2 ; 3)               */
  qsort(clique_set,clique_count,
	sizeof(layout_phase2_cell_t),&layout_phase2_cell_cmp);

  /* Step 3: merge adjacent level with no common host (bottom-up algorithm) */
  DDEBUG("  Step 3: merge adjacent level\n");
  lvl_cur=0;
  if (!(lvl = (layout_level_t *)malloc(sizeof(layout_level_t))))
    FAIL ("Out of memory\n");
  lvl[0].level   = lvl_cur;
  lvl[0].clique  = make_jrb();
  lvl[0].machine = make_jrb();
  lvl_count=1;

  for (i=0;i<clique_count;i++) {
    /* check if there is some common host with previous level */
    for (j=0;j<clique_set[i].clique->count;j++) {
      JRB sm=jrb_find_str(lvl[lvl_cur].machine,
			  clique_set[i].clique->machine[j]->fqdn);
      if (sm) {
	/* Found !! */
	lvl_cur++;
	lvl_count++;
	lvl = (layout_level_t *)realloc(lvl,sizeof(layout_level_t)*lvl_count);
	if (!lvl) FAIL("Out of memory\n");
	lvl[lvl_count - 1].level   = lvl_cur;
	lvl[lvl_count - 1].clique  = make_jrb();
	lvl[lvl_count - 1].machine = make_jrb();

	break; /* found, i said. */
      }
    }
    /* add the machines of this clique to the merged level */
    for (j=0;j<clique_set[i].clique->count;j++) {
      jrb_insert_str(lvl[lvl_cur].machine,
		     clique_set[i].clique->machine[j]->fqdn,
		     new_jval_v(clique_set[i].clique->machine[j]));
    }
    name=clique_set[i].clique->name;
    jrb_insert_str(lvl[lvl_cur].clique,name,new_jval_v(name));
    clique_set[i].level = lvl_cur;
  }

  /* Step 4: We computed a valid level setup. Now, some cosmetic changes:   */
  /*          move all cliques to the level upstairs from were it is having */
  /*          no comon host with the clique. We want to do this change:     */
  /*                                                                        */
  /*  level2:     (A B C D)             \           (A B C D)               */
  /*  level1:     (A E F G)              }-> (A E F G) (B H I J) (C K L)    */
  /*  level0: (B H I J) (C K L) (E M N) /            (E M N)                */
  /*   because cliques (BHIJ) and (CKL) are free to go to level 1           */
  /*                                                                        */
  /*  (top-down algorithm)                                                  */
  DDEBUG("  Step 4: Move up what can be\n");
  for (lvl_cur = lvl_count-2 ; lvl_cur >= 0 ; lvl_cur--) {
    /* for each clique */
    jrb_traverse(sc,lvl[lvl_cur].clique) {
      /* look if we can move it up (ie, no common host with upper level) */
      int found=0;
      c=(clique_t *)jval_v(sc->val);
      if (c->name[0]=='\0') {
	WARN("Traversal too wide\n");
	break;
      }

      //DDEBUG1("< Look at '%s':\n",name);
      while (!found) { /* move the clique as high as possible */
	for (i=0 ; i<c->count ; i++) {
	  if (jrb_find_str(lvl[lvl_cur+1].machine,c->machine[i]->fqdn)) {
	    found = 1;
	    break;
	  }
	}
	if (!found) { /* yuhu, we can move it up */
	  //DDEBUG("  Move it up.\n");
	  /* add the clique to the new level, and remove it from old one */
	  for (i=0; i<c->count; i++) {
	    jrb_insert_str(lvl[lvl_cur + 1].machine, c->machine[i]->fqdn,
			   new_jval_v(c->machine[i]));
	    jrb_delete_node(jrb_find_str(lvl[lvl_cur].machine,
					 c->machine[i]->fqdn));
	  }
	  jrb_insert_str(lvl[lvl_cur + 1].clique, c->name,
			 new_jval_v(c));

	  jrb_delete_node(jrb_find_str(lvl[lvl_cur].clique,c->name));
	}
      }
      //DDEBUG1("> Ok for %s.\n",name);
    }
  }

  /* Step 5: Updates cliques and machines global so that the level of each */
  /*         element is known (top-down algorithm)                         */
  DDEBUG("  Step 5: updates globals\n");
  /* Put all machines to -1 */
  jrb_traverse(sm,machines) {
    m=(machine_t*)jval_v(sm->val);
    if (!m) FAIL("Internal error: can't find the machine in tmp array\n");
    m->z=-1;
  }
  /* do the job */
  for (lvl_cur=lvl_count-1; lvl_cur>=0; lvl_cur--) {    
    jrb_traverse(sc,lvl[lvl_cur].clique) {
      c=(clique_t*)jval_v(sc->val);
      if (!c) FAIL("Internal error: can't find the clique in tmp array\n");
      c->level=lvl_cur;
    }
    jrb_traverse(sm,lvl[lvl_cur].machine) {
      m=(machine_t*)jval_v(sm->val);
      if (m->z<0) {
	m->z=lvl_cur;
      } 
    }   
  }

  free(clique_set);
  return 1;
}

/*----[ layout_phase3 ]------------------------------------------------------*/
/* compute along the X and Y axis                                            */
/*                                                                           */
/* phase 3: Compute the position of hosts (bottom-top-bottom algorithm)      */
/*   bottom-up: set clique member pos relative to the top of the clique.     */
/*     Placement is done on a circle (cone-tree algo).                       */
/*     Each node are allocated some room around this circle:                 */
/*        leaf take 1, roots take (sum of subtree).                          */
/*   top-down: root is {0,0,0}, so all absolute position of each node        */
/*     may be computed                                                       */
/*---------------------------------------------------------------------------*/
static int layout_phase3() {
  JRB sm,sc;
  int lvl_cur;
  machine_t *m,*root;
  clique_t *c;
  int weight_sum;
  int i;

  double theta_inc, theta_cur;
  double min_x,min_y,min_z;
  double max_x,max_y,max_z;
  
  DDEBUG("Layout phase 3\n");
  DDEBUG("  Step 1: bottom-up for relative position\n");
  /* bottom-up: compute the weight of each node, and the relative position   */
  /*            wrt the root of the clique                                   */

  /* Put the weight of all machines to 1 (their own weight) */
  jrb_traverse(sm,machines) {
    m=(machine_t*)jval_v(sm->val);
    m->weight=1;
  }   

  for (lvl_cur = 0 ; lvl_cur < lvl_count ; lvl_cur++) {
    root=NULL;
    jrb_traverse(sc,lvl[lvl_cur].clique) {

      c=(clique_t*)jval_v(sc->val);

      /* Weight of this clique = sum of the weight of its members being at
	 this level*/
      weight_sum=0;
      
      for (i=0; i<c->count; i++) {
	if (c->machine[i]->z== lvl_cur) {
	  weight_sum += c->machine[i]->weight;
	} else {
	  root=c->machine[i];
	}
      }
    
      /* Update the upper level: the node of this clique which are not on this 
	 level are on a upper one. Their weight on this upper level is the
	 weight of this clique (minus one, because the node itself is not
	 weighting at this level)*/
      /* (during the same traversing of the set)
	 compute the x and y of all the nodes which are on this level */
      theta_inc=(3.141592653 * 2) / weight_sum;
      theta_cur=0;
      for (i=0; i<c->count; i++) {
	if (c->machine[i]->z==lvl_cur) {  /* compute the relative x,y */
	  c->machine[i]->father = root;

	  theta_cur += theta_inc * c->machine[i]->weight / 2; /* do half of the way */
	  /* put the point */
	  c->machine[i]->x = RADIUS_MAGNIFICATION * weight_sum * cos(theta_cur);
	  c->machine[i]->y = RADIUS_MAGNIFICATION * weight_sum * sin(theta_cur);
	  theta_cur += theta_inc * c->machine[i]->weight / 2; /* other half of the way */
	} else {
	  /* update weight */
	  c->machine[i]->weight+=((int) (((double)weight_sum) * 
					 WEIGHT_UPPER_LEVEL_MAGNIFICATION));
	}
      }
    }
  }

  /* top-down: Compute absolute position                                     */
  DDEBUG("  Step 2: top-down for absolute position\n");
  for (lvl_cur=lvl_count-2; lvl_cur>=0; lvl_cur--) {    
    jrb_traverse(sm,lvl[lvl_cur].machine) {

      m=(machine_t*)jval_v(sm->val);
    
      if (m && m->father) { /* should allways be the case because root are on the 
			       higher level, which is not explored, but... */
	m->x += m->father->x;
	m->y += m->father->y;
      }
    }
  }

  /* normalize positions (should be >0) */
  DDEBUG("  Step 3: normalize positions (>0 and fitted to display space)\n");
  min_x = min_y = min_z = 0;
  max_x = max_y = max_z = 0;
  jrb_traverse(sm,machines) {
    m = (machine_t *)jval_v(sm->val);
    m->x *= MAGNIFICATION_X;
    m->y *= MAGNIFICATION_Y;
    m->z *= MAGNIFICATION_Z;
    if (m->x < min_x) min_x=m->x;
    if (m->y < min_y) min_y=m->y;
    if (m->z < min_z) min_z=m->z;

    if (m->x > max_x) max_x=m->x;
    if (m->y > max_y) max_y=m->y;
    if (m->z > max_z) max_z=m->z;    
  }
  /*
  DDEBUG5("MAX: X(%.1f - %.1f) Y(%.1f - %.1f) Z(-%.1f)\n",
	  min_x,max_x,	  min_y,max_y,	  max_z);
  DDEBUG3("Scaling factors: X=%f Y=%f Z=%f\n",
	  MAX_X/(max_x-min_x),
	  MAX_Y/(max_y-min_y),
	  MAX_Z/(max_z-min_z));
  */
  jrb_traverse(sm,machines) {
    m = (machine_t *)jval_v(sm->val);
    /* everybody >0, and scaled to fit to display space */
    m->x = MIN(MAX_X, (m->x - min_x) * MAX_X/(max_x-min_x) +1);
    m->y = MIN(MAX_Y, (m->y - min_y) * MAX_Y/(max_y-min_y) +1);
    m->z = MIN(MAX_Z, (m->z - min_z) * MAX_Z/(max_z-min_z) +1);
  }
    

  return 1;
}


/*----[ layout_compute ]-----------------------------------------------------*/
/* Call all phases in order. Kinda main() for layout.                        */
/*---------------------------------------------------------------------------*/
int layout_compute() {
  NWSAPI_HostSpec nsSpec;
  static int fake_clique_nb=6;

  if (!--fake_clique_nb) fake_clique_nb=6;

  if (fake) {
    data_fake(fake_clique_nb,"top","A,B,C,D",
	      "one","A,E,F,G","two","B,H,I,J","three","C,K,L",
	      "four","E,M,N", "five","H,O,P");
  } else {
    nsSpec = 
      *NWSAPI_MakeHostSpec (nameServer,
			    atoi(NWSAPI_EnvironmentValue("NAME_SERVER_PORT", 
							 DEFAULT_NAME_SERVER_PORT)));
    if(!NWSAPI_UseNameServer(&nsSpec))
      ABORT2("Unable to contact name server %s:%d\n",
	     nsSpec.machineName, nsSpec.machinePort);
    /* search the data */
    if (!machines_search() || !cliques_search()) 
      return 0;
  }
  //  cliques_dump(cliques);
  if (!layout_phase1()) return 0;
  if (!layout_phase2()) return 0;
  if (!layout_phase3()) return 0;
  //level_dump();
  //  cliques_dump(cliques);
  level_free();
  //  cliques_dump(cliques);


  return 1;
}
