/* $Id: machines_cliques.c,v 1.2 2002/02/09 23:07:38 mquinson Exp $ */

/* machines_cliques.c - tool functions to handle sets of machines and cliques*/
/* Author: Martin Quinson                                                    */
/* License: LGPL                                                             */

#include <stdio.h>       /* printf() */
#include <stdlib.h>      /* atoi() */
#include <string.h>      /* string functions */
#include <strings.h>     /* strncasecmp() on aix */
#include <unistd.h>      /* getopt() */
#include <stdarg.h>      /* data_fake have a variable number of parameters */

#include "nws_api.h"     /* NWS programming interface */
#include "diagnostic.h"  /* ABORT() */

#include "jval.h"
#include "jrb.h"

#include "machines_cliques.h"

#define DEFAULT_NAME_SERVER_PORT "8090"
#define DEFAULT_SENSOR_PORT      "8060"

JRB machines = NULL;
JRB cliques = NULL;

/*----[ machine_new ]--------------------------------------------------------*/
/* Creates a machine called #name# (If not hostname:port format, 8060 is     */
/*   assumed), and insert it to the global list of machines.                 */
/*---------------------------------------------------------------------------*/
int machine_new(const char *name) {
  char *p;
  machine_t *machine;

  if (!(machine = (machine_t*)malloc(sizeof(machine_t))))
    FAIL("Out of memory\n");
  strcpy(machine->fqdn,name);
  /* split name/port */
  p=strchr(machine->fqdn,':');
  if (p) { 
    *p='\0';
    p++;
  }
  /* get port */
  machine->port= atoi( (p ? p : DEFAULT_SENSOR_PORT) );
  sprintf(machine->series,"%s:%d.availableCpu.0",
	  machine->fqdn,machine->port);
  machine->memberof = 0;
  machine->x = 0;
  machine->y = 0;
  machine->z = 0;
  machine->number = 0;
  machine->father = NULL;

  jrb_insert_str(machines,machine->fqdn,new_jval_v(machine));

  return 1;
}

/*----[ machines_free ]------------------------------------------------------*/
/* Memory leak is a Bad Thing.  (frees the whole set)                        */
/*---------------------------------------------------------------------------*/
void machines_free() {
  JRB sm;
  machine_t *m;

  if (machines) { /* cleanup previous run */
    DDEBUG("Remove all existing machines\n");
    jrb_traverse(sm,machines) {
      m = (machine_t *)jval_v(sm->val);
      free(m);
    }
    jrb_free_tree(machines);
    machines=NULL;
  }
}

/*----[ machines_search ]----------------------------------------------------*/
/* search all machines runing a sensor from the nameserver.                  */
/*---------------------------------------------------------------------------*/
int machines_search() {
  NWSAPI_Object object;
  NWSAPI_ObjectSet filteredObjects;
  NWSAPI_Attribute attribute;

  DDEBUG("Search for machines in NWS\n");

  machines_free();
  machines = make_jrb();

  if(!NWSAPI_GetObjects(NWSAPI_SENSORS, &filteredObjects))
    exit(1);
  
  NWSAPI_ForEachObject(filteredObjects, object) {
    NWSAPI_ForEachAttribute(object, attribute) {
      if(strcmp(NWSAPI_NAME_ATTR, NWSAPI_AttributeName(attribute)) == 0) {
	
	if (!machine_new(NWSAPI_AttributeValue(attribute)))
	  return 0;
      }
    }
  }
  
  NWSAPI_FreeObjectSet(&filteredObjects);
  return 1;
}

/*----[ machines_dump ]------------------------------------------------------*/
/* Dumps all machines for debugging purpose                                  */
/*---------------------------------------------------------------------------*/
int machines_dump(JRB set) {
  JRB tmp;
  machine_t *m;
  int cnt = 0;

  if (!set) {
    printf("(null)\n");
    return 1;
  }
    
  jrb_traverse(tmp,set) {
    m = (machine_t *)jval_v(tmp->val);
    printf("%s:%d -> MembOf %d ; weight %d ; pos=%.0f,%.0f,%.0f\n",
	   m->fqdn,m->port,m->memberof,m->weight,
	   m->x,m->y,m->z);
    cnt++;
  }
  printf("\nmachines_dump: There is %d machines.\n",cnt);

  return 1;
}

/*----[ machines_display ]---------------------------------------------------*/
/* Makes a textual display of the positions:                                 */
/*   Projection on (Ox,Oz) to see z                                          */
/*   Projection on (Ox,Oy) to see x and y                                    */
/*---------------------------------------------------------------------------*/
int machines_display() {
  int Xz=0,Yz=0,Xy=0,Yy=0;
  char clear='.';

#define posz(x,y) (((int)y)*Xz + ((int)x))
#define posy(x,y) (((int)y)*Xy + ((int)x))

  char *oxoz;
  char *oxoy;

  int i,j;

  JRB tmp;
  machine_t *m;
  
  /* how big should be the blackboards ? */
  jrb_traverse(tmp,machines) {
    m = (machine_t *)jval_v(tmp->val);
    if (m->x +1 > Xz)      Xz = m->x + 1;
    if (m->z +2 > Yz)      Yz = m->z + 2;
    if (m->x +1 > Xy)      Xy = m->x + 1;
    if (m->y +1 > Yy)      Yy = m->y + 1;
  }
  Xz++; Yz++;  Xz++; Yz++;
  Xy++; Yy++;  Xy++; Yy++;
  if (!(oxoz=(char*)malloc(sizeof(char)*Xz*Yz+Yz)) ||
      !(oxoy=(char*)malloc(sizeof(char)*Xy*Yy+Yy)))
    FAIL("Out of memory\n");

  /* ....[ oxoz ].... */

  /* erase the blackboard */
  for (i=0;i<Xz;i++)
    for (j=0 ; j<Yz; j++)
      oxoz[posz(i,j)]=clear;

  /* put the hosts */
  jrb_traverse(tmp,machines) {
    m = (machine_t *)jval_v(tmp->val);
    if (oxoz[posz(m->x,m->z +1)] != clear) {
      WARN2("%s erase %c on blackboard oxoz.\n",
	    m->fqdn, oxoz[posz(m->x, m->z +1)]);
    }
    oxoz[posz(m->x, m->z +1)]=m->fqdn[0];
  }
  
  /* display */
  printf ("\n\n+->X\n|\n~\nZ\n   ");
  for (i=0;i<Xz+1;i++) printf("-");  printf("\n");
  for (j=Yz-1;j>=0;j--) {
    oxoz[posz(Xz-1,j)]='\0';
    printf(" %2d|%s|\n",j,oxoz + posz(0,j));
  }
  printf("   ");
  for (i=0;i<Xz+1;i++) printf("-");  printf("\n    ");
  for (i=0;i<Xz;i++) 
    if(i%10 == 0) printf("%d",i/10);
    else printf (" ");
  printf("\n    ");
  for (i=0;i<Xz;i++)  printf("%d",i%10);  printf("\n   ");

  /* ....[ oxoy ].... */

  /* erase the blackboard */
  for (i=0;i<Xy;i++)
    for (j=0 ; j<Yy; j++)
      oxoy[j*Xy + i]=clear;

  /* put the hosts */
  jrb_traverse(tmp,machines) {
    m = (machine_t *)jval_v(tmp->val);
    if (oxoy[posy(m->x, m->y)] != clear) {
      WARN4("%s erase %c on blackboard oxoy (%.0f,%.0f).\n",
	    m->fqdn, oxoy[posy(m->x , m->y)],m->x,m->y);
    }
      oxoy[posy(m->x , m->y)]=m->fqdn[0];
  }

  /* display */
  printf ("\n\nY\n^\n|\n+->X\n    ");

  for (i=0;i<Xy;i++) 
    if(i%10 == 0) printf("%d",i/10);
    else printf (" ");
  printf("\n    ");
  for (i=0;i<Xy;i++)  printf("%d",i%10);  printf("\n   ");

  for (i=0;i<Xy+1;i++) printf("-");  printf("\n");

  for (j=0;j<Yy;j++) {
    oxoy[j*Xy + Xy - 1]='\0';
    printf("%3d|%s|\n",j,oxoy + j*Xy);
  }
  
  printf("   ");
  for (i=0;i<Xy+1;i++) printf("-");  printf("\n    ");

  for (i=0;i<Xy;i++) 
    if(i%10 == 0) printf("%d",(i/10));
    else printf (" ");
  printf("\n    ");
  for (i=0;i<Xy;i++)  printf("%d",i%10);  printf("\n\n\n");

  free (oxoy);
  free (oxoz);
  return 1;
}
/*----[ clique_new ]---------------------------------------------------------*/
/* Creates a clique named #name#, containing #members# (comma separated list)*/
/*  on #layer#, and insert it to the global list of cliques.                 */
/*---------------------------------------------------------------------------*/
int clique_new(char *name,char *members,int layer) {
  clique_t *clique;
  machine_t *machine;
  JRB sm;
  char *word;

  if (!(clique = (clique_t*)malloc(sizeof(clique_t))))
    FAIL("Out of memory\n");

  strcpy(clique->name,name);
  clique->count = 0;
  clique->machine=NULL;
  clique->level=0;
  for(word = strtok(members, ",");
      (word != NULL);
      word = strtok(NULL, ",")) {
    sm=jrb_find_str(machines, word);
    if (!sm) {
      char *p=word;
      while (*p != '\0' && *p != ':') p++;
      *p='\0';
      if (!machine_new(word))
	return 0;
      if (!(sm=jrb_find_str(machines, word)))
	FAIL("Damnit, I can find the machine I just added\n");      
    }
    machine=(machine_t*)jval_v(sm->val);
    clique->count++;
    if (clique->machine) {
      clique->machine=(machine_t**)realloc(clique->machine,sizeof(char*)*clique->count);
    } else {
      clique->machine=(machine_t**)malloc(sizeof(char*));
    }
    if (!clique->machine) 
      FAIL("Out of memory\n");
    
    clique->machine[ clique->count-1 ] = machine;
  }
  jrb_insert_str(cliques,clique->name,new_jval_v(clique));
  
  return 1;
}
/*----[ cliques_free ]-------------------------------------------------------*/
/* Memory leak is a Bad Thing.  (frees the whole set)                        */
/*---------------------------------------------------------------------------*/
void cliques_free() {
  JRB sc;
  clique_t *c;

  if (cliques) { /* cleanup previous run */
    jrb_traverse(sc,cliques) {
      c = (clique_t *)jval_v(sc->val);
      free(c->machine);
      free(c);
    }
    jrb_free_tree(cliques);
    cliques=NULL;
  }
}


/*----[ cliques_search ]-----------------------------------------------------*/
/* search all cliques runing a sensor from the nameserver.                   */
/*---------------------------------------------------------------------------*/
int cliques_search() {
  NWSAPI_Object object;
  NWSAPI_ObjectSet filteredObjects;
  NWSAPI_Attribute attribute;

  JRB sc; /* search clique */
  char *name=NULL,*members=NULL;

  cliques_free();
  cliques = make_jrb();

  DDEBUG("Search for cliques in NWS\n");

  if(!NWSAPI_GetObjects(NWSAPI_CLIQUES, &filteredObjects))
    exit(1);
  
  NWSAPI_ForEachObject(filteredObjects, object) {
    NWSAPI_ForEachAttribute(object, attribute) {
      if(strcmp(NWSAPI_NAME_ATTR, NWSAPI_AttributeName(attribute)) == 0) {
	if (!(name=strdup(NWSAPI_AttributeValue(attribute))))
	  FAIL("Out of memory\n");
      } else if(strcmp(NWSAPI_MEMBER_ATTR, NWSAPI_AttributeName(attribute)) == 0) {
	if (!(members=strdup(NWSAPI_AttributeValue(attribute))))
	  FAIL("Out of memory\n");
      }
    }
    if (!name || !members)
      FAIL("Weird ! Name or members not defined for this clique.\n");
    
    sc=jrb_find_str(cliques,name);
    if (sc) {
      WARN1("The clique %s is reported twice. Ignore the subsequent one.\n",name);
    } else {
      clique_new(name,members,0);
    }
    free(name);
    free(members);
    name = members = NULL;
  } 
  
  NWSAPI_FreeObjectSet(&filteredObjects);
  return 1;
}

/*----[ cliques_dump ]-------------------------------------------------------*/
/* Dumps all cliques for debugging purpose                                   */
/*---------------------------------------------------------------------------*/
int cliques_dump(JRB set) {
  JRB tmp;
  clique_t *c;
  int cnt = 0;
  int i;

  if (!set) {
    printf("(null)\n");
    return 1;
  }
    
  jrb_traverse(tmp,set) {
    c = (clique_t *)jval_v(tmp->val);
    printf(">> %s Count=%d;Level=%d;Member=(",c->name,c->count,c->level);
    for (i=0 ; i<c->count ; i++) {
      printf("%s%s",c->machine[i]->fqdn,
	     (i<c->count-1 ? "," : ")\n"));
    }
    cnt++;
  }
  printf("\ncliques_dump: There is %d cliques.\n",cnt);

  return 1;
}

/*----[ data_fake ]----------------------------------------------------------*/
/* Fakes data with #nbcliques#, each one been described with a subsequent    */
/*  string (debuging purpose. Strings are coma (,) separated:                */
/*    The first word there is the name of the clique.                        */
/*    The next ones are name of machines in the clique.                      */
/* This is caller responsability to make sure that there is enough string for*/
/*  #nbcliques#.                                                             */
/*---------------------------------------------------------------------------*/
int data_fake(int nbcliques, ...) {
  char *desc,*name;
  int clique_num;

  va_list pa;

  machines = make_jrb();
  cliques = make_jrb();

  va_start(pa,nbcliques);
  for (clique_num = 0 ; clique_num<nbcliques ; clique_num++) {
    name=va_arg(pa,char*);
    desc=strdup(va_arg(pa,char*)); /* HACK: Little leak here, but this string
				            have to be writable for strtok */
    DDEBUG2("Fake clique '%s': '%s'\n",name,desc);
    
    if (!clique_new(name,desc,0)) 
      return 0;
    free(desc);
  }

  va_end(pa);

  return 1;
}
