/* $Id: render.c,v 1.4 2002/02/09 23:08:50 mquinson Exp $ */

/* render - functions to actually render the stuff                           */
/* Author: Martin Quinson                                                    */
/* License: LGPL                                                             */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cichlid_server.h" 
#include "utils.h"            /* to compute the color to render the load */
/* FIXME: utils.h actually contains also functions. Please move wherelse */

#include "diagnostic.h"       /* ABORT() */
#include "jval.h"
#include "jrb.h"

#include "machines_cliques.h" /* machines and cliques sets */
#include "settings.h"         /* constants influing the layout computation */
#include "layout.h"           /* compute hosts' place in visualization space */

#include "render.h"

extern JRB machines;
extern JRB cliques;

static struct DataSet *DSet = NULL;
static struct VtxInfo vertices[MAX_VERTICES];
static struct EdgeInfo edges[MAX_EDGES];
static const float rect_space[3] = {1000,1000,1000};

static int vertex_count=0,edge_count=0; /* host / link */
static int StartTime, LastCalc;
extern int fake;

/* defines the kind of objects we may want to render */
struct SiteAttribs {
    VtxDraw style;
    float size;
    unsigned char color[3];
};
static const struct SiteAttribs SiteAttr[1] = {
  {VD_SPHERE, 40, {0,255,0}}
};

struct LinkAttribs {
    EdgeDraw style;
    float size;
    unsigned char color[3];
    float speed;
};
static const struct LinkAttribs LinkAttr[1] = {
  {ER_CYLINDER, 10, {0,20,255}, 5.0} // Link, only one kind for now
};

/*----[ render_init ]--------------------------------------------------------*/
/* Make what is needed prior to the rendering.                               */
/*---------------------------------------------------------------------------*/
int render_init() {
  JRB tmp;
  machine_t *m;
  clique_t *c;
  int i,j;

  struct VtxInfo vertex_empty = {{0.0, 0.0, 0.0}, VD_SPHERE, 0, {0,0,0}};
  struct EdgeInfo edge_empty = {{0, 0}, ER_CYLINDER, ER_NONE, 0, {0,0,0}};
  
  /*....[ Cleanup ].............................*/
  for (i=0; i<MAX_VERTICES; i++) {
    vertices[i] = vertex_empty;
    veset_set_vlabel(DSet, i, NULL, NULL);
  }
  for (i=0; i<MAX_EDGES; i++) {
    edges[i] = edge_empty;
  }
  if (vertex_count)
    veset_set_vtcs(DSet, 0, &vertices[0], vertex_count);
  if (edge_count)
    veset_set_edges(DSet, 0, &edges[0], edge_count);

  /*....[ compute the needed dimensions ].......*/
  vertex_count = edge_count = 0;
  jrb_traverse(tmp,machines) {
    m = (machine_t *)jval_v(tmp->val);
    m->number = ++vertex_count;
  }
  vertex_count ++; /* vertex 0 is empty */

  jrb_traverse(tmp,cliques) {
    c = (clique_t *)jval_v(tmp->val);
    edge_count += (c->count * (c->count-1)) / 2;
    DDEBUG3("edge_count %d after %s (c->count=%d)\n",
	    edge_count,c->name,c->count);
  }

  /*....[ Validity check ]......................*/
  if (vertex_count >= MAX_VERTICES) 
    FAIL2("Sorry, I can only handle %d vertices (you asked for %d). "
	  "Change MAX_VERTICES in settings.h and recompile me if you want more\n",
	  MAX_VERTICES,vertex_count);
  if (edge_count >= MAX_EDGES) 
    FAIL2("Sorry, I can only handle %d edges (you asked for %d). "
	  "Change MAX_EDGES in settings.h and recompile me if you want more\n",
	  MAX_EDGES,edge_count);

  DDEBUG2("vertex_count %d edge_count %d\n",vertex_count,edge_count);

  /*....[ Compute new objects ]..................*/
  jrb_traverse(tmp,machines) {
    m = (machine_t *)jval_v(tmp->val);
    if (m->z<0) {
      WARN("FIXME: Why have some machine a gost on negative Z ? (fixitdonthideit)\n");
      continue;
    }
    vertices[m->number].loc[0] = m->x;
    vertices[m->number].loc[1] = m->y;
    vertices[m->number].loc[2] = m->z;
    vertices[m->number].style   =  SiteAttr[0].style;
    vertices[m->number].size    =  SiteAttr[0].size;
    vertices[m->number].color[0] = SiteAttr[0].color[0];
    vertices[m->number].color[1] = SiteAttr[0].color[1];
    vertices[m->number].color[2] = SiteAttr[0].color[2];

    i=m->number;
    //DDEBUG5("%s (%d) => %.0f %.0f %.0f\n",m->fqdn,i,
    //	    vertices[i].loc[0],vertices[i].loc[1],vertices[i].loc[2]);
  }

  edge_count = 0;
  jrb_traverse(tmp,cliques) {
    c=(clique_t*)jval_v(tmp->val);
    for (i=0; i<c->count; i++) {
      for (j=i+1; j<c->count; j++) {
	edges[edge_count].ends[0] = c->machine[i]->number;
	edges[edge_count].ends[1] = c->machine[j]->number;

	//	DDEBUG5("Link %d: %s (%d) <-> %s (%d)\n",edge_count,
	//	c->machine[i]->fqdn, c->machine[i]->number,
	//	c->machine[j]->fqdn, c->machine[j]->number);

        edges[edge_count].direction = ED_NONE;
        edges[edge_count].style = LinkAttr[0].style;
        edges[edge_count].size  = LinkAttr[0].size;
        edges[edge_count].color[0] = LinkAttr[0].color[0];
        edges[edge_count].color[1] = LinkAttr[0].color[1];
        edges[edge_count].color[2] = LinkAttr[0].color[2];
	edge_count++;
      }
    }
  }

  /*....[ Put new objects in place ]............*/
  veset_set_vtcs(DSet, 0, &vertices[0], vertex_count);
  veset_set_edges(DSet, 0, &edges[0], edge_count);

  /*....[ Set labels ]..........................x*/
  jrb_traverse(tmp,machines) {
    char *p;
    m = (machine_t *)jval_v(tmp->val);
    if (m->z < 0)
      continue;
    
    strcpy(m->sname,m->fqdn);
    p=strchr(m->sname,'.');
    if (p)
      *p='\0';
    
    veset_set_vlabel(DSet, m->number, m->sname, NULL);
  }
  for (i=0;i<=edge_count;i++) {
    veset_set_elabel(DSet, i, NULL, NULL);
  }

  LastCalc = sys_time_get();
  return 1;
}
/*----[ render_reset ]-------------------------------------------------------*/
/* What to do when someone connects.                                         */
/*---------------------------------------------------------------------------*/
static void render_reset(const char *remote) {
  LastCalc = sys_time_get();
  //  veset_set_vtcs(DSet, 0, &vertices[0], vertex_count);
}

/*----[ render_update ]------------------------------------------------------*/
/* Get updates from NWS.                                                     */
/*---------------------------------------------------------------------------*/
static void render_update(void) {
  static int until_redo_layout=COMPUTE_LAYOUT_EVERY_N_STEP;
  int t, slice_count;
  /*JRB tmp;
    machine_t *m;*/
  
  t = sys_time_get() - LastCalc;
  slice_count = t / SLICE_MSEC;
  if (slice_count < 1) {
    sys_time_sleep(SLICE_MSEC - t);
    slice_count = 1;
    t = SLICE_MSEC;
  }
  
  LastCalc = sys_time_get();
  
  veset_set_edges(DSet, 0, &edges[0], edge_count);
  veset_set_vtcs(DSet, 0, &vertices[0], vertex_count);
  //  veset_dump_vtcs(DSet,1,vertex_count);
  //  veset_dump_edges(DSet,1,edge_count);

  if (!until_redo_layout--) {
    until_redo_layout=COMPUTE_LAYOUT_EVERY_N_STEP;
    DDEBUG(" -- Layout update  --\n");
    if (!layout_compute()  || !render_init())
      exit(1);
  }    
  if (!fake) {
    DDEBUG(" -- Render_update (TODO) --\n");
    // FIXME: do something usefull
    
    /*
      jrb_traverse(tmp,machines) {
      m = (machine_t *)jval_v(tmp->val);
      
      DDEBUG5("%s (%d) -> %.0f %.0f %.0f\n",m->sname,m->number,
      vertices[m->number].loc[0],vertices[m->number].loc[1],
      vertices[m->number].loc[2]);
      }
    */
  }
}

/*----[ render_point ]-------------------------------------------------------*/
/* What to do when the pointer goes over an object                           */
/*---------------------------------------------------------------------------*/


/*----[ render ]-------------------------------------------------------------*/
/* render the stuff. Does not return (if no error), it's a main loop.        */
/*---------------------------------------------------------------------------*/
int render() {
  DDEBUG("Begin rendering\n");
  DSet = veset_new("NWS visualization tool",0,MAX_VERTICES,MAX_EDGES,
		   rect_space,CS_POSRECT);
  if (!DSet) 
    FAIL("Damnit ! veset_new returned NULL\n");
  dsSetDataSet(DSet);

  if (!render_init()) return 0;

  veset_dump_vtcs(DSet,1,vertex_count);
  veset_dump_edges(DSet,1,edge_count);
  dsConnectFunc(render_reset);
  dsDataFunc(render_update);
  dsMainLoop();

  return 1; /* never reached */
}
